@@ -1,5 +1,5 @@
 /* board.cc -- game board for falling blocks game.  -*- C++ -*-
-   Copyright (C) 2011-2015 Gerardo Ballabio
+   Copyright (C) 2011-2020 Gerardo Ballabio
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,9 +21,12 @@
 #include "grid.h"
 #include "group.h"
 
-bool game_board::fits(int b, int s, coords c) const
+#include <iostream>
+#include <iomanip>
+
+bool game_board::fits(const block &b, int s, coords c) const
 {
-  std::vector<coords> bricks = get_bricks(blocks[b].shape[s]);
+  std::vector<coords> bricks = get_bricks(b.shape[s]);
   for (std::vector<coords>::const_iterator i=bricks.begin();
        i!=bricks.end(); ++i)
     {
@@ -35,9 +38,9 @@ bool game_board::fits(int b, int s, coords c) const
   return true;
 }
 
-bool game_board::fits_b(int b, int s, coords c) const
+bool game_board::fits_b(const block &b, int s, coords c) const
 {
-  std::vector<coords> bricks = get_bricks(blocks[b].blocker[s]);
+  std::vector<coords> bricks = get_bricks(b.blocker[s]);
   for (std::vector<coords>::const_iterator i=bricks.begin();
        i!=bricks.end(); ++i)
     {
@@ -49,27 +52,27 @@ bool game_board::fits_b(int b, int s, coords c) const
   return true;
 }
 
-coords game_board::move_in(int b, int s, coords c) const
+coords game_board::move_in(const block &b, int s, coords c) const
 {
   // move down if necessary
-  while (blocks[b].shape[s].get_y0() + c.y < cells.get_y0())
+  while (b.shape[s].get_y0() + c.y < cells.get_y0())
     c = g->down(c);
 
   // move right if necessary
-  while (blocks[b].shape[s].get_x0() + c.x < cells.get_x0())
+  while (b.shape[s].get_x0() + c.x < cells.get_x0())
     c = g->right(c);
 
   // move left if necessary
-  while (blocks[b].shape[s].get_x0() + blocks[b].shape[s].get_cols() + c.x >
+  while (b.shape[s].get_x0() + b.shape[s].get_cols() + c.x >
 	 cells.get_x0() + cells.get_cols())
     c = g->left(c);
 
   // move back if necessary
-  while (blocks[b].shape[s].get_z0() + c.z < cells.get_z0())
+  while (b.shape[s].get_z0() + c.z < cells.get_z0())
     c = g->back(c);
 
   // move front if necessary
-  while (blocks[b].shape[s].get_z0() + blocks[b].shape[s].get_layers() + c.z >
+  while (b.shape[s].get_z0() + b.shape[s].get_layers() + c.z >
 	 cells.get_z0() + cells.get_layers())
     c = g->front(c);
 
@@ -78,18 +81,14 @@ coords game_board::move_in(int b, int s, coords c) const
 
 void game_board::first_block()
 {
-  // initialize random numbers generator
-  Glib::TimeVal timeval;
-  timeval.assign_current_time();
-  std::srand(static_cast<unsigned>(timeval.as_double() * G_USEC_PER_SEC));
-
   nsize = bmax;
   // choose block
-  nblock = rand() % series[nsize];
+  nindex = rand() % series[nsize];
   for (int i=0; i<nsize; ++i)
-    nblock += series[i];
+    nindex += series[i];
+  nblock = blocks[nindex];
   // choose orientation
-  nshape = rand() % blocks[nblock].shape.size();
+  nshape = rand() % nblock.shape.size();
 
   // move to active block
   next_block();
@@ -98,12 +97,12 @@ void game_board::first_block()
 void game_board::land_block()
 {
   // dump block onto cells grid
-  std::vector<coords> bricks = get_bricks(blocks[iblock].shape[ishape]);
+  std::vector<coords> bricks = get_bricks(fblock.shape[fshape]);
   for (std::vector<coords>::const_iterator i=bricks.begin();
        i!=bricks.end(); ++i)
     {
       coords n = pos + *i;
-      cells.set(n, iblock);
+      cells.set(n, findex);
     }
   score += score_block();
 
@@ -124,38 +123,91 @@ void game_board::next_level()
     case 1:
       // increase maximum block size
       ++bmax;
-      next_blockset();
       switch (mode2)
 	{
 	case 0: if (bmin > 1) --bmin; break;
 	case 1: break;
 	case 2: ++bmin; break;
 	}
+      start_blockset();
       break;
     case 2:
       // increase superblock size
       ++bsuper;
-      next_blockset();
+      start_blockset();
       break;
     }
 }
 
 void game_board::next_blockset()
 {
-  // extract last blockset
-  int first = blocks.size() - ((series.size() > 0) ? series.back() : 0);
-  std::vector<block> v(blocks.begin() + first, blocks.end());
+  if (blocks.size() == 0)
+    {
+      blocks.push_back(g->block_zero());
+      series.push_back(1);
+      return;
+    }
 
-  // reduce it to a maximum size in order to speed up generation
-  int max = g->blockset_max();
-  if (static_cast<int>(v.size()) > max)
-    v = random_blocks(v, max);
+  // extract last blockset
+  unsigned offset = blocks.size() - ((series.size() > 0) ? series.back() : 0);
+  std::vector<unsigned> last(blocks.size() - offset);
+  for (unsigned i=0; i<last.size(); ++i)
+    last[i] = offset + i;
+
+  // limit size in order to speed up generation
+  unsigned max1 = g->blockset_max(), max2 = max1 * 5;
+  if (max2 > last.size())
+    max2 = last.size();
+
+  if (max1 < last.size())
+    // randomize
+    for (unsigned i=0; i<max2; ++i)
+      {
+	unsigned index = i + (rand() % (last.size() - i));
+	if (index > i)
+	  std::swap(last[index], last[i]);
+      }
 
   // generate and append new blockset
-  v = g->augment_blocks(v, cells.get_x0(), cells.get_z0(),
-			cells.get_cols(), cells.get_layers());
-  blocks.insert(blocks.end(), v.begin(), v.end());
-  series.push_back(v.size());
+  std::vector<block> v;
+  for (unsigned i=0; i<max2 && (i<max1 || !stop_blockset); ++i)
+    g->augment_block(blocks[last[i]], v, cells.get_x0(), cells.get_z0(),
+		     cells.get_cols(), cells.get_layers());
+
+  // blocks, series may be reallocated
+  // so this must be safeguarded to prevent the main thread from copying an
+  // invalid block
+  {
+    // acquire lock
+    std::lock_guard<std::mutex> lock(mutex);
+
+    blocks.insert(blocks.end(), v.begin(), v.end());
+    series.push_back(v.size());
+
+    // lock released at end of scope
+  }
+}
+
+void game_board::start_blockset()
+{
+  // stop worker if it hasn't already
+  end_blockset();
+
+  // offload blockset generation to a separate thread
+  stop_blockset = false;
+  worker = new std::thread(&game_board::next_blockset, this);
+}
+
+void game_board::end_blockset()
+{
+  // terminate blockset-generating thread
+  stop_blockset = true;
+  if (worker)
+    {
+      worker->join();
+      delete worker;
+      worker = 0;
+    }
 }
 
 int game_board::score_block() const
@@ -168,9 +220,22 @@ int game_board::score_lines(int lines) const
   return score_block() * 20 * (1 << lines);
 }
 
+game_board::~game_board()
+{
+  if (g)
+    delete g;
+
+  if (worker)
+    {
+      worker->join();
+      delete worker;
+      worker = 0;
+    }
+}
+
 grid<bool> game_board::get_block() const
 {
-  grid<bool> g = blocks[iblock].shape[ishape];
+  grid<bool> g = fblock.shape[fshape];
   g.move(pos);
   return g;
 }
@@ -179,10 +244,10 @@ grid<bool> game_board::get_shadow()
 {
   // move block down until it lands
   coords c = pos;
-  while (fits(iblock, ishape, g->down(c)))
+  while (fits(fblock, fshape, g->down(c)))
     c = g->down(c);
 
-  grid<bool> g = blocks[iblock].shape[ishape];
+  grid<bool> g = fblock.shape[fshape];
   g.move(c);
   return g;
 }
@@ -197,6 +262,11 @@ void game_board::start_game(const group *gg, int width, int depth,
   if (g)
     delete g;
 
+  // initialize random numbers generator
+  Glib::TimeVal timeval;
+  timeval.assign_current_time();
+  std::srand(static_cast<unsigned>(timeval.as_double() * G_USEC_PER_SEC));
+
   // set spatial group
   g = gg;
   // generate game board
@@ -205,6 +275,7 @@ void game_board::start_game(const group *gg, int width, int depth,
   std::vector<block> v;
   blocks.resize(0);
   series.resize(0);
+  stop_blockset = false;
   for (int i=0; i<=max; ++i)
     next_blockset();
   if (min > max)
@@ -227,45 +298,50 @@ void game_board::start_game(const group *gg, int width, int depth,
   filled.resize(0);
   game = true;
 
+  // more blocksets needed
+  if (mode1 == 1 || mode1 == 2)
+    start_blockset();
+
   first_block();
 }
 
 bool game_board::rotate(int n)
 {
-  int rotations = blocks[iblock].rotate.size() / blocks[iblock].shape.size();
+  int rotations = fblock.rotate.size() / fblock.shape.size();
   if (n >= rotations)
     return false; // do nothing
 
-  int nshape = blocks[iblock].rotate[ishape * rotations + n];
-  coords m = move_in(iblock, nshape, pos);
-  if (!fits(iblock, nshape, m))
+  int nshape = fblock.rotate[fshape * rotations + n];
+  coords m = move_in(fblock, nshape, pos);
+  if (!fits(fblock, nshape, m))
     return false; // reject move
 
   // accept move
   pos = m;
-  ishape = nshape;
+  fshape = nshape;
   return true;
 }
 
 bool game_board::reflect()
 {
-  int nshape = blocks[iblock].reflect[ishape];
-  coords m = move_in(iblock, nshape, pos);
-  if (!fits(iblock, nshape, m))
+  int nshape = fblock.reflect[fshape];
+  coords m = move_in(fblock, nshape, pos);
+  if (!fits(fblock, nshape, m))
     return false; // reject move
 
   // accept move
   pos = m;
-  ishape = nshape;
+  fshape = nshape;
   return true;
 }
 
 bool game_board::left()
 {
   coords c = g->left(pos);
-  if (!fits(iblock, ishape, c))
-    return false;
+  if (!fits(fblock, fshape, c))
+    return false; // reject move
 
+  // accept move
   pos = c;
   return true;
 }
@@ -273,9 +349,10 @@ bool game_board::left()
 bool game_board::right()
 {
   coords c = g->right(pos);
-  if (!fits(iblock, ishape, c))
-    return false;
+  if (!fits(fblock, fshape, c))
+    return false; // reject move
 
+  // accept move
   pos = c;
   return true;
 }
@@ -283,9 +360,10 @@ bool game_board::right()
 bool game_board::front()
 {
   coords c = g->front(pos);
-  if (!fits(iblock, ishape, c))
-    return false;
+  if (!fits(fblock, fshape, c))
+    return false; // reject move
 
+  // accept move
   pos = c;
   return true;
 }
@@ -293,9 +371,10 @@ bool game_board::front()
 bool game_board::back()
 {
   coords c = g->back(pos);
-  if (!fits(iblock, ishape, c))
-    return false;
+  if (!fits(fblock, fshape, c))
+    return false; // reject move
 
+  // accept move
   pos = c;
   return true;
 }
@@ -303,7 +382,7 @@ bool game_board::back()
 bool game_board::down()
 {
   coords c = g->down(pos);
-  if (fits(iblock, ishape, c) && fits_b(iblock, ishape, pos))
+  if (fits(fblock, fshape, c) && fits_b(fblock, fshape, pos))
     {
       // move block down and continue
       pos = c;
@@ -327,7 +406,6 @@ bool game_board::drop()
 
 void game_board::color_lines(int c)
 {
-  // remove them
   for (std::vector<int>::const_iterator i=filled.begin();
        i!=filled.end(); ++i)
     g->color_line(cells, *i, c);
@@ -354,16 +432,22 @@ void game_board::remove_lines()
 void game_board::next_block()
 {
   // new active block
-  iblock = nblock;
-  ishape = nshape;
+  fblock = nblock;
+  findex = nindex;
+  fshape = nshape;
   bsize = nsize;
   // set initial position
   pos = make_coords();
-  if (iblock != 0)
-    pos = move_in(iblock, ishape, pos);
+  if (findex != 0)
+    pos = move_in(fblock, fshape, pos);
   // if it doesn't fit, game is over
-  if (!fits(iblock, ishape, pos))
-    game = false;
+  if (!fits(fblock, fshape, pos))
+    {
+      // stop worker if it hasn't already
+      end_blockset();
+
+      game = false;
+    }
 
   // next block
   int max_super = bsuper;
@@ -393,12 +477,22 @@ void game_board::next_block()
       if (nsize < bmin)
 	nsize += (plus > 0) ? plus : 1;
     }
-  // choose block
-  nblock = rand() % series[nsize];
-  for (int i=0; i<nsize; ++i)
-    nblock += series[i];
+  // this must be safeguarded because blocks, series may be reallocated
+  // in the worker thread
+  {
+    // acquire lock
+    std::lock_guard<std::mutex> lock(mutex);
+
+    // choose block
+    nindex = rand() % series[nsize];
+    for (int i=0; i<nsize; ++i)
+      nindex += series[i];
+    nblock = blocks[nindex];
+
+    // lock released at end of scope
+  }
   // choose orientation
-  nshape = rand() % blocks[nblock].shape.size();
+  nshape = rand() % nblock.shape.size();
 }
 
 void game_board::stop()
@@ -406,6 +500,9 @@ void game_board::stop()
   if (!game)
     return;
 
+  // stop worker if it hasn't already
+  end_blockset();
+
   game = false;
   cells = grid<int>(cells.get_rows(), cells.get_cols(), cells.get_layers(),
 		    cells.get_c0());
