@@ -1,5 +1,5 @@
 /* group.h -- spatial groups for falling block game.  -*- C++ -*-
-   Copyright (C) 2011-2015 Gerardo Ballabio
+   Copyright (C) 2011-2020 Gerardo Ballabio
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,28 +24,27 @@
 struct block
 {
   std::vector<grid<bool> > shape, blocker;
-  std::vector<int> rotate, reflect;
+  std::vector<unsigned> rotate, reflect;
 };
 
-std::vector<block> random_blocks(const std::vector<block> &v, int n);
-
 // virtual base class
 class group
 {
   // abstract interface
 private:
-  virtual int size() const = 0;
-  virtual int rotations() const = 0;
-  virtual int rotate(int n, int m) const = 0;
-  virtual int reflect(int n) const = 0;
+  // define symmetry group, translations etc.
+  virtual unsigned size() const = 0;
+  virtual unsigned rotations() const = 0;
+  virtual unsigned rotate(unsigned n, unsigned m) const = 0;
+  virtual unsigned reflect(unsigned n) const = 0;
   virtual int type(const coords &c) const = 0;
   virtual std::vector<coords> neighbors(const coords &c) const = 0;
   virtual std::vector<coords> blockers(const coords &c) const = 0;
-  virtual grid<bool> transform(const grid<bool> &g, int n) const = 0;
+  virtual grid<bool> transform(const grid<bool> &g, unsigned n) const = 0;
   virtual grid<bool> center(const grid<bool> &g) const = 0;
 public:
   virtual bool is_3d() const = 0;
-  virtual int blockset_max() const = 0; // values determined empirically
+  virtual unsigned blockset_max() const = 0; // values determined empirically
 
   // game
   virtual coords left(const coords &c) const = 0;
@@ -80,15 +79,16 @@ private:
   grid<bool> blockers_set(const grid<bool> &g) const;
   block make_block(const grid<bool> &g) const;
 public:
-  std::vector<block> augment_blocks(const std::vector<block> &v,
-				    int x0 = 0, int z0 = 0,
-				    int cols = 1, int layers = 1) const;
+  block block_zero() const;
+  void augment_block(const block &b, std::vector<block> &v,
+		     int x0 = 0, int z0 = 0,
+		     int cols = 1, int layers = 1) const;
 };
 
 class group_2d : public group
 {
 private:
-  virtual int rotations() const { return 1; }
+  virtual unsigned rotations() const { return 1; }
   virtual coords front(const coords &c) const { return c; }
   virtual coords back(const coords &c) const { return c; }
 public:
@@ -98,7 +98,7 @@ public:
 class group_3d : public group
 {
 private:
-  virtual int rotations() const { return 3; }
+  virtual unsigned rotations() const { return 3; }
 public:
   virtual bool is_3d() const { return true; }
   virtual void draw_board_front(Cairo::RefPtr<Cairo::Context> context,
